import logging
from abc import ABCMeta
from typing import Any, Dict, List, Optional

import datayoga_core.blocks.redis.utils as redis_utils
import redis
from datayoga_core import expression
from datayoga_core.block import Block as DyBlock
from datayoga_core.connection import Connection
from datayoga_core.context import Context
from datayoga_core.result import BlockResult, Result, Status

logger = logging.getLogger("dy")


class Block(DyBlock, metaclass=ABCMeta):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")

        connection_details = Connection.get_connection_details(self.properties["connection"], context)
        self.redis_client = redis_utils.get_client(connection_details)

        self.command = self.properties.get("command", "HSET")

        key = self.properties["key"]
        self.key_expression = expression.compile(key["language"], key["expression"])

        logger.info(f"Writing to Redis connection '{self.properties.get('connection')}'")

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        pipeline = self.redis_client.pipeline()
        block_result = BlockResult()
        for record in data:
            # transform to a list, filtering out None, which Redis does not support
            dict_as_list = sum(filter(
                lambda i: i[1] is not None and not i[0].startswith(Block.INTERNAL_FIELD_PREFIX),
                record.items()
            ), ())
            pipeline.execute_command(self.command, self.key_expression.search(record), *dict_as_list)

        try:
            results = pipeline.execute(raise_on_error=False)
            for record, result in zip(data, results):
                if isinstance(result, Exception):
                    block_result.rejected.append(Result(Status.REJECTED, message=f"{result}", payload=record))
                else:
                    block_result.processed.append(Result(Status.SUCCESS, payload=record))
        except redis.exceptions.ConnectionError as e:
            raise ConnectionError(e)

        return block_result



import logging
from abc import ABCMeta
from typing import Any, Dict, List, Optional

import cassandra.auth
from cassandra.cluster import NoHostAvailable, PreparedStatement
from datayoga_core import write_utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.connection import Connection
from datayoga_core.context import Context
from datayoga_core.opcode import OpCode
from datayoga_core.result import BlockResult, Result, Status

logger = logging.getLogger("dy")


class Block(DyBlock, metaclass=ABCMeta):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")

        connection_name = self.properties["connection"]
        connection_details = Connection.get_connection_details(connection_name, context)
        if connection_details.get("type") != "cassandra":
            raise ValueError(f"{connection_name} is not a cassandra connection")

        auth_provider = cassandra.auth.PlainTextAuthProvider(
            username=connection_details.get("user"),
            password=connection_details.get("password"))

        self.cluster = cassandra.cluster.Cluster(
            connection_details.get("hosts"),
            port=connection_details.get("port", 9042),
            auth_provider=auth_provider)

        self.keyspace = self.properties.get("keyspace")

        logger.debug(f"Connecting to Cassandra")
        self.session = self.cluster.connect(self.keyspace)

        self.opcode_field = self.properties.get("opcode_field")
        self.table = self.properties.get("table")
        self.keys = self.properties.get("keys")
        self.mapping = self.properties.get("mapping")

        business_key_columns = [column["column"] for column in write_utils.get_column_mapping(self.keys)]
        mapping_columns = [column["column"] for column in write_utils.get_column_mapping(self.mapping)]

        pk_clause = f"{' and '.join([column + ' = ?' for column in business_key_columns])}"
        self.delete_stmt = f"delete from {self.keyspace}.{self.table} where {pk_clause}"

        self.upsert_stmt = f"update {self.keyspace}.{self.table} set {', '.join([column + ' = ?' for column in mapping_columns])} where {pk_clause}"

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        logger.debug(f"Running {self.get_block_name()}")

        opcode_groups = write_utils.group_records_by_opcode(data, opcode_field=self.opcode_field)
        # reject any records with unknown or missing Opcode
        rejected_records: List[Result] = []

        for opcode in set(opcode_groups.keys()) - {o.value for o in OpCode}:
            rejected_records.extend([
                Result(status=Status.REJECTED, payload=record, message=f"unknown opcode '{opcode}'")
                for record in opcode_groups[opcode]
            ])

        try:
            self.execute_upsert(opcode_groups[OpCode.CREATE] + opcode_groups[OpCode.UPDATE])
            self.execute_delete(opcode_groups[OpCode.DELETE])
        except NoHostAvailable as e:
            raise ConnectionError(e)

        return BlockResult(
            processed=[Result(Status.SUCCESS, payload=record)
                       for opcode in OpCode for record in opcode_groups[opcode.value]],
            rejected=rejected_records)

    def get_future(self, stmt: PreparedStatement, record: Dict[str, Any]) -> Any:
        future = self.session.execute_async(stmt, record)
        future.add_errback(lambda ex, record: Result(status=Status.REJECTED, payload=record, message=f"{ex}"), record)
        return future

    def execute_upsert(self, records: List[Dict[str, Any]]):
        if records:
            logger.debug(f"Upserting {len(records)} record(s) to {self.table} table")
            stmt = self.session.prepare(self.upsert_stmt)
            futures = []
            for record in records:
                record_to_upsert = write_utils.map_record(record, self.keys, self.mapping)
                futures.append(self.get_future(stmt, record_to_upsert))

            for future in futures:
                future.result()

    def execute_delete(self, records: List[Dict[str, Any]]):
        if records:
            logger.debug(f"Deleting {len(records)} record(s) from {self.table} table")
            stmt = self.session.prepare(self.delete_stmt)
            futures = []
            for record in records:
                record_to_delete = write_utils.map_record(record, self.keys)
                futures.append(self.get_future(stmt, record_to_delete))

            for future in futures:
                future.result()

    def stop(self):
        self.cluster.shutdown()



import logging
from abc import ABCMeta
from typing import Any, Dict, List, Optional

import jinja2
from datayoga_core import utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.context import Context
from datayoga_core.result import BlockResult, Result, Status

logger = logging.getLogger("dy")


class Block(DyBlock, metaclass=ABCMeta):
    field: str
    template: jinja2.Template

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")

        self.field = self.properties["field"]
        self.template = jinja2.Template(self.properties["template"])

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        logger.debug(f"Running {self.get_block_name()}")

        block_result = BlockResult()

        for row in data:
            try:
                # assign the new values
                utils.set_field(row, self.field, self.template.render(**row))
                block_result.processed.append(Result(Status.SUCCESS, payload=row))
            except Exception as e:
                block_result.rejected.append(Result(status=Status.REJECTED, payload=row, message=f"{e}"))

        return block_result



import pytest
from datayoga_core import result, utils
from datayoga_core.blocks.filter.block import Block


@pytest.mark.asyncio
async def test_filter_sql():
    block = Block(
        {
            "language": "sql",
            "expression": "age>20"
        }
    )
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe", "age": 25}]) == utils.all_success([{"fname": "john", "lname": "doe", "age": 25}])


@pytest.mark.asyncio
async def test_filter_sql_multiple():
    block = Block(
        {
            "language": "sql",
            "expression": "age>20"
        }
    )
    block.init()
    assert tuple(await block.run(
        [
            {"fname": "john", "lname": "doe", "age": 25},
            {"fname": "john2", "lname": "doe2", "age": 15},
            {"fname": "john3", "lname": "doe3", "age": 22}
        ]
    )) == (
        [
            result.Result(status=result.Status.SUCCESS, payload={"fname": "john", "lname": "doe", "age": 25}),
            result.Result(status=result.Status.SUCCESS, payload={"fname": "john3", "lname": "doe3", "age": 22})
        ],
        [
            result.Result(status=result.Status.FILTERED, payload={"fname": "john2", "lname": "doe2", "age": 15})
        ],
        []
    )


@pytest.mark.asyncio
async def test_filter_jmespath_multiple():
    block = Block(
        {
            "language": "jmespath",
            "expression": "age>`20`"
        }
    )
    block.init()
    assert tuple(await block.run(
        [
            {"fname": "john", "lname": "doe", "age": 25},
            {"fname": "john2", "lname": "doe2", "age": 15},
            {"fname": "john3", "lname": "doe3", "age": 22}
        ]
    )) == (
        [
            result.Result(status=result.Status.SUCCESS, payload={"fname": "john", "lname": "doe", "age": 25}),
            result.Result(status=result.Status.SUCCESS, payload={"fname": "john3", "lname": "doe3", "age": 22})
        ],
        [
            result.Result(status=result.Status.FILTERED, payload={"fname": "john2", "lname": "doe2", "age": 15})
        ],
        []
    )


@pytest.mark.asyncio
async def test_filter_sql_not():
    block = Block(
        {
            "language": "sql",
            "expression": "age>20"
        }
    )
    block.init()
    assert tuple(await block.run([{"fname": "john", "lname": "doe", "age": 15}])) == ([], [result.Result(status=result.Status.FILTERED, payload={"fname": "john", "lname": "doe", "age": 15})], [])


@pytest.mark.asyncio
async def test_filter_jmespath():
    block = Block(
        {
            "language": "jmespath",
            "expression": "age>`20`"
        }
    )
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe", "age": 25}]) == utils.all_success([{"fname": "john", "lname": "doe", "age": 25}])


@pytest.mark.asyncio
async def test_filter_jmespath_multiple_nested():
    block = Block(
        {
            "language": "jmespath",
            "expression": "age.years>`20`"
        }
    )
    block.init()
    assert tuple(await block.run(
        [
            {"fname": "john", "lname": "doe", "age": {"years": 25, "months": 5}},
            {"fname": "john2", "lname": "doe2", "age": {"years": 15, "months": 1}},
            {"fname": "john3", "lname": "doe3", "age": {"years": 22, "months": 2}}
        ]
    )) == (
        [
            result.Result(status=result.Status.SUCCESS, payload={"fname": "john", "lname": "doe", "age": {"years": 25, "months": 5}}),
            result.Result(status=result.Status.SUCCESS, payload={"fname": "john3", "lname": "doe3", "age": {"years": 22, "months": 2}})
        ],
        [
            result.Result(status=result.Status.FILTERED, payload={"fname": "john2", "lname": "doe2", "age": {"years": 15, "months": 1}})
        ],
        []
    )


@pytest.mark.asyncio
async def test_filter_jmespath_not():
    block = Block(
        {
            "language": "jmespath",
            "expression": "age>`20`"
        }
    )
    block.init()
    assert tuple(await block.run([{"fname": "john", "lname": "doe", "age": 15}])) == ([], [result.Result(status=result.Status.FILTERED, payload={"fname": "john", "lname": "doe", "age": 15})], [])



from prometheus_client import Counter, start_http_server

incoming_records = Counter("incoming_records", "Number of incoming records")
processed_entries = Counter("processed_records", "Number of processed records", ("step",))
rejected_records = Counter("rejected_records", "Number of rejected records", ("step",))
filtered_records = Counter("filtered_records", "Number of filtered records", ("step",))


def start(port: int):
    """Starts the prometheus metrics exporter on selected port."""

    start_http_server(port)



import ssl
from typing import Any, Dict

import redis
from redis import Redis


def get_client(connection_details: Dict[str, Any]) -> Redis:
    """Establishes a connection to a Redis server with optional SSL/TLS encryption and authentication.

    Args:
        connection_details (Dict[str, Any]): A dictionary containing connection parameters:
            - "host" (str): The Redis server hostname or IP address.
            - "port" (int): The Redis server port number.
            - "user" (Optional[str]): Redis username.
            - "password" (Optional[str]): Redis password.
            - "key" (Optional[str]): Path to the client private key file for SSL/TLS.
            - "key_password" (Optional[str]): Password for the client private key file.
            - "cert" (Optional[str]): Path to the client certificate file for SSL/TLS.
            - "cacert" (Optional[str]): Path to the CA certificate file for SSL/TLS.

    Returns:
        Redis: Redis client instance.

    Raises:
        ValueError: If connection to Redis fails.
    """
    if connection_details["type"] != "redis":
        raise ValueError("not a Redis connection")

    host = connection_details["host"]
    port = connection_details["port"]
    key = connection_details.get("key")
    key_password = connection_details.get("key_password")
    cert = connection_details.get("cert")
    cacert = connection_details.get("cacert")

    try:
        client = redis.Redis(
            host=host,
            port=port,
            username=connection_details.get("user"),
            password=connection_details.get("password"),
            ssl=(key is not None and cert is not None) or cacert is not None,
            ssl_keyfile=key,
            ssl_password=key_password,
            ssl_certfile=cert,
            ssl_cert_reqs=ssl.CERT_REQUIRED if cacert else ssl.CERT_NONE,
            ssl_ca_certs=cacert,
            decode_responses=True,
            client_name="datayoga"
        )
        client.json()

        client.ping()
        return client
    except Exception as e:
        raise ValueError(f"can not connect to Redis on {host}:{port}:\n {e}")



import pytest
from datayoga_core import utils
from datayoga_core.blocks.add_field.block import Block


@pytest.mark.asyncio
async def test_add_field():
    """Test case for adding a field using JMESPath expression."""
    block = Block({
        "field": "full_name",
        "language": "jmespath",
        "expression": "[fname, lname] | join(' ', @)"
    })
    block.init()

    assert await block.run([
        {"fname": "john", "lname": "doe"}
    ]) == utils.all_success([
        {"fname": "john", "lname": "doe", "full_name": "john doe"}
    ])


@pytest.mark.asyncio
async def test_add_nested_field():
    """Test case for adding a nested field using JMESPath expression."""
    block = Block({
        "field": "name.full_name",
        "language": "jmespath",
        "expression": "[name.fname, name.lname] | join(' ', @)"
    })
    block.init()

    assert await block.run([
        {"name": {"fname": "john", "lname": "doe"}}
    ]) == utils.all_success([
        {"name": {"fname": "john", "lname": "doe", "full_name": "john doe"}}
    ])


@pytest.mark.asyncio
async def test_add_nested_field_parent_key_missing():
    """Test case for adding a nested field when the parent key is missing."""
    block = Block({
        "field": "location.street",
        "language": "jmespath",
        "expression": "lname"
    })
    block.init()

    assert await block.run([
        {
            "fname": "jane",
            "lname": "smith",
            "country_code": 1,
            "country_name": "usa",
            "credit_card": "1234-5678-0000-9999",
            "gender": "F"
        }
    ]) == utils.all_success([
        {
            "fname": "jane",
            "lname": "smith",
            "country_code": 1,
            "country_name": "usa",
            "credit_card": "1234-5678-0000-9999",
            "location": {"street": "smith"},
            "gender": "F"
        }
    ])


@pytest.mark.asyncio
async def test_add_multiple_fields():
    """Test case for adding multiple fields using JMESPath expressions."""
    block = Block({
        "fields": [
            {
                "field": "name.full_name",
                "language": "jmespath",
                "expression": "concat([name.fname, ' ', name.lname])"
            },
            {
                "field": "name.fname_upper",
                "language": "jmespath",
                "expression": "upper(name.fname)"
            }
        ]
    })
    block.init()

    assert await block.run([
        {
            "name": {
                "fname": "john",
                "lname": "doe"
            }
        }
    ]) == utils.all_success([
        {
            "name": {
                "fname": "john",
                "lname": "doe",
                "full_name": "john doe",
                "fname_upper": "JOHN"
            }
        }
    ])


@pytest.mark.asyncio
async def test_add_field_with_dot():
    """Test case for adding a field with a dot in the name using JMESPath expression."""
    block = Block({
        "field": "name\.full_name",
        "language": "jmespath",
        "expression": "[fname, lname] | join(' ', @)"
    })
    block.init()

    assert await block.run([
        {"fname": "john", "lname": "doe"}
    ]) == utils.all_success([
        {"fname": "john", "lname": "doe", "name.full_name": "john doe"}
    ])



import copy
import os
import re
import sys
import uuid
from os import path
from typing import Any, Dict, List, Union

import orjson
import yaml
from datayoga_core import result
from datayoga_core.block import Block
from datayoga_core.expression import JMESPathExpression, Language
from datayoga_core.result import BlockResult, Result, Status


def read_json(filename: str) -> Any:
    """Loads a filename as a JSON object.

    Args:
        filename (str): JSON filename to load.

    Returns:
        Any: JSON object.
    """
    with open(filename, "r", encoding="utf8") as json_file:
        return orjson.loads(json_file.read())


def read_yaml(filename: str) -> Dict[str, Any]:
    """Loads a filename as a YAML object.

    Args:
        filename (str): YAML filename to load.

    Raises:
        ValueError: In case of invalid YAML.

    Returns:
         Dict[str, Any]: YAML python object.
    """
    try:
        with open(filename, "r", encoding="utf8") as yaml_file:
            return yaml.safe_load(yaml_file)
    except Exception as e:
        raise ValueError(f"Malformed YAML: {e}")


def format_block_properties(properties: Dict[str, Any]) -> Dict[str, Any]:
    """Adds `fields` array with the passed properties in case it's missing.

    Args:
        properties (Dict[str, Any]): properties.

    Returns:
        Dict[str, Any]: formatted properties with `fields` array.
    """
    return {"fields": [properties]} if not "fields" in properties else properties


def prepare_expression(language: Language, expr: Union[Dict, str]) -> str:
    """Prepares the expression to compile.

    The SQL expression will be dumped as JSON.
    For JMESPath, the expression will be generated.

    Args:
        language (Language): The language for which the expression is prepared (SQL or JMESPath).
        expr (Union[Dict, str]): The expression to prepare.

    Raises:
        ValueError: If the map expression is not in a JSON-like format enclosed in { }.

    Returns:
        str: The prepared expression.
    """
    def prepare_key(key: str) -> str:
        return f'"{key}"' if " " in key or "-" in key else key

    def prepare_value(value):
        if isinstance(value, dict):
            return prepare_expression(language, value)

        return value

    if language == Language.SQL:
        return orjson.dumps(expr).decode() if isinstance(expr, dict) else expr.strip()

    # If there is an object here, we generate an expression.
    if isinstance(expr, dict):
        expr = ", ".join(f'{prepare_key(k)}: {prepare_value(v)}' for k, v in expr.items())
        expr = f"{{{expr}}}"

    expression_prop = expr.strip()

    if not (expression_prop.startswith("{") and expression_prop.endswith("}")):
        raise ValueError("map expression must be in a json-like format enclosed in { }")

    return expression_prop


def is_bundled() -> bool:
    return getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")


def get_bundled_dir() -> str:
    datayoga_dir = path.join(sys._MEIPASS, "datayoga_core")
    return datayoga_dir if os.path.isdir(datayoga_dir) else sys._MEIPASS


def get_resource_path(relative_path: str) -> str:
    if is_bundled():
        # we are running in a bundle
        return path.join(get_bundled_dir(), "resources", relative_path)
    else:
        # we are running in a normal Python environment
        return path.join(os.path.dirname(__file__), "resources", relative_path)


def set_field(obj: Dict[str, Any], field_name: str, value: Any):
    """Sets the value of a field in a nested dictionary.

    Args:
        obj (Dict): The dictionary to modify.
        field_name (str): Field name representing the path to the field, e.g., "nested.field".
        value (Any): The value to set.

    Example:
        set_field(obj, "nested.field", 42) will set obj["nested"]["field"] = 42.
    """
    field_path = split_field(field_name)
    for key in field_path[:-1]:
        key = unescape_field(key)
        obj = obj.setdefault(key, {})  # Setdefault creates missing nested keys as dictionaries

    obj[unescape_field(field_path[-1])] = value


def split_field(field: str, *, __expression=re.compile(r"(?<!\\)\.")) -> List[str]:
    r"""Splits a string by dots, similar to str.split(), but allows escaping a dot with a backslash.
    Example:
        a.b\.c.d => ["a", "b\\.c", "d"]
    """
    return __expression.split(field)


def unescape_field(field: str) -> str:
    r"""Replaces "\\." by "."."""
    return field.replace("\\.", ".")


def all_success(records: List[Dict[str, Any]]) -> BlockResult:
    return BlockResult(processed=[Result(Status.SUCCESS, payload=row) for row in records])


def is_rejected(record: Dict[str, Any]) -> bool:
    return record.get(Block.RESULT_FIELD, result.SUCCESS).status == Status.REJECTED


def add_uid(record: Dict[str, Any]) -> Dict[str, Any]:
    return {Block.MSG_ID_FIELD: f"{uuid.uuid4()}", **record}


def remove_msg_id(record: dict) -> dict:
    return {k: v for k, v in record.items() if k != Block.MSG_ID_FIELD}


def explode_records(records: List[Dict[str, Any]], field_expression: str) -> List[Dict[str, Any]]:
    """Takes a list of records and a JMESPath expression specifying a field to be exploded, and returns a new list of records
    where each record has been "exploded" into multiple records based on the values in the specified field.

    Args:
        records (List[Dict[str, Any]]): A list of dictionaries representing records.
        field_expression (str): A string specifying the field to be exploded, in the form "field_name: expression".

    Returns:
        List[Dict[str, Any]]: A new list of dictionaries representing the exploded records.
    """
    field_name, expression = map(str.strip, field_expression.split(":", maxsplit=1))

    jmespath_expr = JMESPathExpression()
    jmespath_expr.compile(expression)

    exploded_records = []

    for record in records:
        # Apply the JMESPath expression to the current record to obtain the values to be exploded.
        field_values = jmespath_expr.search(record)

        # If the JMESPath expression returned any values, create a new record for each value and add it to the output list.
        if field_values:
            for field_value in field_values:
                new_record = copy.deepcopy(record)
                new_record[field_name] = field_value
                exploded_records.append(new_record)

    return exploded_records



from __future__ import annotations

import asyncio
import logging
import marshal
import os
import sys
from contextlib import suppress
from enum import Enum, unique
from pathlib import Path
from typing import Any, Dict, List, Optional

import jsonschema
from datayoga_core import blocks, prometheus, utils
from datayoga_core.block import Block
from datayoga_core.context import Context
from datayoga_core.producer import Producer
from datayoga_core.result import JobResult, Result, Status
from datayoga_core.step import Step

logger = logging.getLogger("dy")


@unique
class ErrorHandling(str, Enum):
    ABORT = "abort"
    IGNORE = "ignore"


class Job:
    """Job.

    Attributes:
        steps List[Block]: List of steps.
    """

    def __init__(self, steps: Optional[List[Step]] = None, producer: Optional[Producer] = None,
                 error_handling: Optional[ErrorHandling] = None):
        """Constructs a job and its blocks.

        Args:
            steps (List[Dict[str, Any]]): Job steps.
            producer (Optional[Producer]): Block to be used as a producer.
            error_handling (Optional[ErrorHandling]): error handling strategy.
        """
        self.producer = producer
        self.steps = steps
        self.error_handling = error_handling if error_handling else ErrorHandling.IGNORE
        self.initialized = False
        self.root = None

    def init(self, context: Optional[Context] = None):
        # open any connections or setup needed
        self.context = context

        if self.steps is not None and len(self.steps) > 0:
            # set the first step as root
            self.root = self.steps[0]
            self.root.init(context)
            self.root.add_done_callback(self.handle_results)
            last_step = self.root

            # create the step sequence
            for step in self.steps[1:]:
                step.init(context)

                last_step = last_step.append(step)

        if self.producer:
            self.producer.init(context)

        self.initialized = True

    def transform(self, data: List[Dict[str, Any]], deepcopy: bool = True) -> JobResult:
        """Transforms data.

        Args:
            data (List[Dict[str, Any]]): Data.
            deepcopy (bool): If set to True, performs a deepcopy before modifying records;
                             otherwise, it modifies them in place. This can affect performance.

        Returns:
            JobResult: Job result.
        """
        if not self.initialized:
            logger.debug("job has not been initialized yet, initializing...")
            self.init()
        # use marshal. faster than deepcopy
        transformed_data = marshal.loads(marshal.dumps(data)) if deepcopy else data

        result = JobResult()
        # create an event loop for the duration of the transformation
        loop = asyncio.new_event_loop()

        try:
            for step in self.steps:
                try:
                    if len(transformed_data) == 0:
                        # in case all records have been filtered, stop sending
                        break
                    processed, filtered, rejected = loop.run_until_complete(step.block.run(transformed_data))
                    result.filtered.extend(filtered)
                    result.rejected.extend(rejected)
                    transformed_data = [result.payload for result in processed]
                except ConnectionError as e:
                    # connection errors are thrown back to the caller to handle
                    raise e
                except Exception as e:
                    # other exceptions are rejected
                    logger.error(f"Error while transforming data: {e}")
                    result.rejected.extend(
                        [Result(Status.REJECTED, payload=row, message=f"{e}") for row in transformed_data])
                    return result

            # the processed records are those that make it to the end
            result.processed = [Result(Status.SUCCESS, payload=row) for row in transformed_data]
        finally:
            # close the event loop
            with suppress(NotImplementedError):
                # for pyodide. doesn't implement loop.close, ignore
                loop.close()

        return result

    async def async_transform(self, data: List[Dict[str, Any]], deepcopy: bool = True) -> JobResult:
        """Transforms data (async version).

        Args:
            data (List[Dict[str, Any]]): Data.
            deepcopy (bool): If set to True, performs a deepcopy before modifying records;
                             otherwise, it modifies them in place. This can affect performance.

        Returns:
            JobResult: Job result.
        """
        if not self.initialized:
            logger.debug("job has not been initialized yet, initializing...")
            self.init()
        # use marshal. faster than deepcopy
        transformed_data = marshal.loads(marshal.dumps(data)) if deepcopy else data

        result = JobResult()

        for step in self.steps:
            try:
                if len(transformed_data) == 0:
                    # in case all records have been filtered, stop sending
                    break
                processed, filtered, rejected = await step.block.run(transformed_data)
                result.filtered.extend(filtered)
                result.rejected.extend(rejected)
                transformed_data = [result.payload for result in processed]
            except ConnectionError as e:
                # connection errors are thrown back to the caller to handle
                raise e
            except Exception as e:
                # other exceptions are rejected
                logger.error(f"Error while transforming data: {e}")
                result.rejected.extend(
                    [Result(Status.REJECTED, payload=row, message=f"{e}") for row in transformed_data])
                return result

        # the processed records are those that make it to the end
        result.processed = [Result(Status.SUCCESS, payload=row) for row in transformed_data]

        return result

    async def run(self):
        async for records in self.producer.produce():
            prometheus.incoming_records.inc(len(records))

            logger.debug(f"Retrieved records:\n\t{records}")
            await self.root.process(records)

        await self.shutdown()

    async def shutdown(self):
        # wait for in-flight records to finish
        await self.root.join()

        # graceful shutdown
        await self.root.stop()

    def handle_results(self, msg_ids: List[str], results: List[Result]):
        if any(x.status == Status.REJECTED for x in results) and self.error_handling == ErrorHandling.ABORT:
            logger.critical("Aborting due to rejected record(s)")
            sys.exit(1)

        self.producer.ack(msg_ids)

    @staticmethod
    def validate(source: Dict[str, Any], whitelisted_blocks: Optional[List[str]] = None):
        jsonschema.validate(instance=source, schema=Job.get_json_schema(whitelisted_blocks))

    @staticmethod
    def compile(source: Dict[str, Any], whitelisted_blocks: Optional[List[str]] = None) -> Job:
        Job.validate(source, whitelisted_blocks=whitelisted_blocks)
        steps: List[Step] = []
        # parse the steps
        for step_definition in source.get("steps"):
            block_type = step_definition.get("uses")
            block: Block = Block.create(block_type, step_definition.get("with"))
            step: Step = Step(step_definition.get("id", block_type), block)
            steps.append(step)

        # parse the input
        input_block = None
        if source.get("input") is not None:
            input_definition = source.get("input")
            input_block = Block.create(input_definition.get("uses"), input_definition.get("with"))

        return Job(steps, input_block, source.get("error_handling"))

    @staticmethod
    def get_json_schema(whitelisted_blocks: Optional[List[str]] = None) -> Dict[str, Any]:
        """"Compiles a complete json schema of the job and all possible blocks"""
        block_schemas = []
        # we traverse the json schemas directly instead of 'walk_packages'
        # to avoid importing all of the block classes
        blocks_dir = (os.path.join(utils.get_bundled_dir(), "blocks") if utils.is_bundled() else
                      os.path.dirname(os.path.realpath(blocks.__file__)))
        block_types = []
        for schema_path in Path(blocks_dir).rglob("**/block.schema.json"):
            block_type = os.path.relpath(
                os.path.dirname(schema_path),
                os.path.dirname(os.path.realpath(blocks.__file__))
            )
            block_type = block_type.replace(os.path.sep, ".")

            if not (whitelisted_blocks is not None and block_type not in whitelisted_blocks):
                block_types.append(block_type)
                # load schema file
                schema = utils.read_json(f"{schema_path}")
                # append to the array of allOf for the full schema
                # we use allOf for better error reporting
                block_schemas.append({
                    "if": {
                        "properties": {
                            "uses": {
                                "description": "Block type",
                                "type": "string",
                                "const": block_type
                            },
                        },
                        "required": ["uses"]
                    },
                    "then": {
                        "properties": {
                            "with": schema,
                        }
                    }
                })

        job_schema = utils.read_json(
            os.path.join(
                utils.get_bundled_dir() if utils.is_bundled() else os.path.dirname(os.path.realpath(__file__)),
                "resources", "schemas", "job.schema.json"))
        job_schema["definitions"]["block"] = {
            "type": "object",
            "properties": {
                "uses": {
                    "enum": block_types,
                }
            },
            "allOf": block_schemas
        }

        return job_schema



import logging
from abc import ABCMeta
from typing import Any, Dict, List, Optional

from datayoga_core import expression, utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.context import Context
from datayoga_core.expression import Expression
from datayoga_core.result import BlockResult, Result, Status

logger = logging.getLogger("dy")


class Block(DyBlock, metaclass=ABCMeta):
    expression: Expression

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")

        language = self.properties["language"]
        expression_prop = utils.prepare_expression(language, self.properties["expression"])

        self.expression = expression.compile(language, expression_prop)

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        logger.debug(f"Running {self.get_block_name()}")
        mapped_rows = self.expression.search_bulk(data)
        # we always add the internal fields back, in case they were dropped in the map operation
        for mapped_row, original_row in zip(mapped_rows, data):
            original_msg_id = original_row.get(Block.MSG_ID_FIELD)
            original_opcode = original_row.get(Block.OPCODE_FIELD)
            if original_msg_id is not None:
                mapped_row[Block.MSG_ID_FIELD] = original_msg_id
            if original_opcode is not None:
                mapped_row[Block.OPCODE_FIELD] = original_opcode

        return BlockResult(processed=[Result(Status.SUCCESS, payload=row) for row in mapped_rows])



from abc import abstractmethod
from typing import Any, AsyncGenerator, Dict, List

from .block import Block


class Message:
    def __init__(self, msg_id: str, value: Dict[str, Any]):
        self.msg_id = msg_id
        self.value = value


class Producer(Block):

    @abstractmethod
    async def produce(self) -> AsyncGenerator[List[Message], None]:
        """Produces data

        Returns:
            AsyncGenerator[List[Message], None]: A generator of message batches.
        """
        raise NotImplementedError

    def ack(self, msg_ids: List[str]):
        """Sends acknowledge for the message IDs of the records that have been processed

        Args:
            msg_ids (List[str]): Message IDs
        """
        pass



import logging
from abc import ABCMeta
from typing import Any, Dict, List, Optional

from datayoga_core import expression
from datayoga_core.block import Block as DyBlock
from datayoga_core.context import Context
from datayoga_core.result import BlockResult, Result, Status

logger = logging.getLogger("dy")


class Block(DyBlock, metaclass=ABCMeta):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        self.expression = expression.compile(self.properties["language"], self.properties["expression"])

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        result = BlockResult()
        logger.debug(f"Running {self.get_block_name()}")
        return_data = self.expression.filter(data, tombstone=True)
        # mark filtered rows
        for i, row in enumerate(return_data):
            if row is None:
                result.filtered.append(Result(Status.FILTERED, payload=data[i]))
            else:
                result.processed.append(Result(Status.SUCCESS, payload=row))

        return result



import logging
import os
from abc import ABCMeta
from typing import Any, Dict, List, Optional

from datayoga_core import utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.context import Context
from datayoga_core.result import BlockResult
from datayoga_core.utils import remove_msg_id
from fastparquet import write
from pandas import DataFrame

logger = logging.getLogger("dy")


class Block(DyBlock, metaclass=ABCMeta):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        parquet_file = self.properties["file"]

        if os.path.isabs(parquet_file) or context is None:
            self.file = parquet_file
        else:
            self.file = os.path.join(context.properties.get("data_path"), parquet_file)

        logger.debug(f"file: {self.file}")

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        logger.debug("Writing parquet")

        out = [{"data": remove_msg_id(record)} for record in data]
        append = os.path.exists(self.file)
        write(self.file, DataFrame(out), append=append, has_nulls=True)

        # if we made it here, it is a success. return the data and the success result
        return utils.all_success(data)



import logging
from abc import ABCMeta
from asyncio import Queue
from contextlib import suppress
from itertools import count
from typing import AsyncGenerator, List, Optional

import orjson
from aiohttp.web import (BaseRequest, HTTPInternalServerError, HTTPOk,
                         Response, Server, ServerRunner, TCPSite)
from datayoga_core.context import Context
from datayoga_core.producer import Message
from datayoga_core.producer import Producer as DyProducer

logger = logging.getLogger("dy")


class Block(DyProducer, metaclass=ABCMeta):
    port: int
    host: str

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        self.port = int(self.properties.get("port", 8080))
        self.host = self.properties.get("host", "0.0.0.0")

    async def produce(self) -> AsyncGenerator[List[Message], None]:
        queue = Queue(maxsize=1000)

        async def handler(request: BaseRequest) -> Response:
            try:
                queue.put_nowait(orjson.loads(await request.read()))
                return HTTPOk()
            except Exception:  # noqa
                logger.exception("Got exception while parsing request:")
                return HTTPInternalServerError()

        runner = ServerRunner(Server(handler))
        await runner.setup()
        srv = TCPSite(runner, self.host, self.port)
        await srv.start()
        logger.info(f"Listening on {self.host}:{self.port}...")

        try:
            counter = iter(count())

            while True:
                data = await queue.get()
                yield [{self.MSG_ID_FIELD: f"{next(counter)}", **data}]

        finally:
            with suppress(Exception):
                await srv.stop()



import logging
from abc import ABCMeta
from typing import Any, Dict, List, Optional

import datayoga_core.blocks.redis.utils as redis_utils
import redis
from datayoga_core import expression, utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.connection import Connection
from datayoga_core.context import Context
from datayoga_core.result import BlockResult, Result, Status

logger = logging.getLogger("dy")


class Block(DyBlock, metaclass=ABCMeta):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")

        connection_details = Connection.get_connection_details(self.properties.get("connection"), context)

        # Dry mode is internal and used for validate the block without establishing a connection.
        # This behavior should be implemented in a common way, see this issue: https://lnk.pw/eklj
        if not self.properties.get("dry"):
            self.redis_client = redis_utils.get_client(connection_details)

        self.field_path = [utils.unescape_field(field) for field in utils.split_field(self.properties.get("field"))]

        self.cmd = self.properties["cmd"]
        args = self.properties["args"]
        self.args_expressions = [expression.compile(self.properties["language"], c) for c in args]

        logger.info(f"Using Redis connection '{self.properties.get('connection')}'")

    async def run(self, data: List[Dict[str, Any]]) -> BlockResult:
        logger.debug(f"Running {self.get_block_name()}")

        pipeline = self.redis_client.pipeline(transaction=False)
        block_result = BlockResult()

        for record in data:
            params = [self.cmd]
            for expr in (c.search(record) for c in self.args_expressions):
                params.extend(expr if isinstance(expr, list) else [expr])

            pipeline.execute_command(*params)

        try:
            results = pipeline.execute(raise_on_error=False)
            for record, result in zip(data, results):
                if isinstance(result, Exception):
                    block_result.rejected.append(Result(Status.REJECTED, message=f"{result}", payload=record))
                    continue

                obj = record
                for field in self.field_path[:-1]:
                    obj = obj.setdefault(field, {})

                obj[self.field_path[-1]] = result

                block_result.processed.append(Result(Status.SUCCESS, payload=record))
        except redis.exceptions.ConnectionError as expr:
            raise ConnectionError(expr)

        return block_result



import json
from typing import Generator

import pytest
from aioresponses import aioresponses
from datayoga_core import Context, utils
from datayoga_core.blocks.http.write.block import Block


@pytest.fixture
def mock_aioresponse() -> Generator[aioresponses, None, None]:
    """Fixture providing an aioresponses instance for mocking aiohttp requests.

    Yields:
        aioresponses: An instance of aioresponses for mocking aiohttp requests.
    """
    with aioresponses() as m:
        yield m


@pytest.mark.asyncio
async def test_http_write(mock_aioresponse: aioresponses):
    """Test case for the HTTP write block with mocked requests.put."""
    context = Context({
        "connections": {
            "http_example": {
                "type": "http",
                "base_uri": "https://datayoga.io",
                "headers": {
                        "Authorization": {
                            "expression": "concat(['Bearer ', credit_card])",
                            "language": "jmespath"
                        },
                    "Content-Type": "application/json"
                },
                "query_parameters": {
                    "country": {
                        "expression": "country_name",
                        "language": "jmespath"
                    },
                    "origin_country": "israel"
                },
                "timeout": 12
            }
        }
    })

    block = Block({
        "connection": "http_example",
        "endpoint": {
            "expression": "concat(['users/', id])",
            "language": "jmespath"
        },
        "method": "PUT",
        "payload": {
            "full_name": {
                "expression": "concat([fname, ' ' , lname])",
                "language": "jmespath"
            },
        },
        "extra_headers": {
            "custom_header": {
                "expression": "lname || '-' || fname",
                "language": "sql"
            }
        },
        "extra_query_parameters": {
            "fname": {
                "expression": "UPPER(fname)",
                "language": "sql"
            }
        },
        "output": {
            "status_code": "response.status_code",
            "headers": "response.headers",
            "body": "response.content"
        },
        "timeout": 3
    })

    block.init(context)

    url = "https://datayoga.io/users/123?country=usa&fname=JOHN&origin_country=israel"

    mock_aioresponse.put(url,
                         status=200,
                         body=json.dumps({"status": "success"}),
                         headers={"my_header": "123", 'Content-Type': 'application/json'})

    assert await block.run([
        {
            "id": 123,
            "credit_card": "1234-5678-0000-9999",
            "fname": "john",
            "lname": "doe",
            "country_name": "usa"
        }
    ]) == utils.all_success([
        {
            "id": 123,
            "credit_card": "1234-5678-0000-9999",
            "fname": "john",
            "lname": "doe",
            "country_name": "usa",
            "response": {
                "status_code": 200,
                "headers": {
                    "Content-Type": "application/json",
                    "my_header": "123"
                },
                "content": '{"status": "success"}'
            }
        }
    ])

    # Validate the aiohttp request
    calls = mock_aioresponse.requests

    assert len(calls) == 1

    request_key = list(calls.keys())[0]
    request = calls[request_key][0].kwargs

    assert request_key[0] == "PUT"
    assert f"{request_key[1]}" == url

    # Validate the request headers
    assert request["headers"]["Authorization"] == "Bearer 1234-5678-0000-9999"
    assert request["headers"]["Content-Type"] == "application/json"
    assert request["headers"]["custom_header"] == "doe-john"

    # Validate the request query parameters
    assert request["params"]["country"] == "usa"
    assert request["params"]["origin_country"] == "israel"
    assert request["params"]["fname"] == "JOHN"

    # Validate the request body (data)
    assert request["data"] == {"full_name": "john doe"}
    assert request["timeout"] == 3



import asyncio
import logging

from datayoga_core.step import Step

logger = logging.getLogger("dy")


class StepBuffer(Step):
    def __init__(self, step_id: str, min_buffer_size=4, max_buffer_size=4, flush_ms=1000):
        super().__init__(step_id, None)
        self.min_buffer_size = min_buffer_size
        self.max_buffer_size = max(max_buffer_size, min_buffer_size)
        self.buffer = []
        self.flush_ms = flush_ms
        self.timer = None
        self.concurrency_lock = asyncio.Semaphore(1)

    async def flush_timer(self):
        await asyncio.sleep(self.flush_ms/1000)
        logger.debug("flushing on timeout")
        await self.flush()

    async def run(self, worker_id: int):
        while True:
            entry = await self.queue.get()
            logger.debug(f"appending {entry}")
            if (self.timer is None or self.timer.cancelled()) and self.flush_ms is not None:
                # first record, we add a timer
                logger.debug("creating timer")
                self.timer = asyncio.create_task(self.flush_timer())
            else:
                logger.debug("no timer")
            self.buffer.extend(entry)

            if len(self.buffer) >= self.min_buffer_size:
                logger.debug(f"flushing on buffer size {self.buffer} {len(self.buffer)}")
                if self.timer:
                    self.timer.cancel()
                await self.flush()

    async def flush(self):
        # flush buffer
        await self.concurrency_lock.acquire()
        try:
            # we may have accumulated a larger buffer while we flushed, so flush in max_buffer_size batches
            while len(self.buffer) > 0:
                logging.debug(f"flushing {self.buffer}")
                # check if we have a next step
                if self.next_step:
                    # process downstream
                    logging.debug(f"sending to next step")
                    await self.next_step.process([self.buffer.pop(0) for _ in range(min(len(self.buffer), self.max_buffer_size))])
        finally:
            self.concurrency_lock.release()



import logging
from enum import Enum, unique
from typing import Tuple

import sqlalchemy as sa
from datayoga_core.connection import Connection
from datayoga_core.context import Context

logger = logging.getLogger("dy")


@unique
class DbType(str, Enum):
    DB2 = "db2"
    MYSQL = "mysql"
    ORACLE = "oracle"
    PSQL = "postgresql"
    SQLSERVER = "sqlserver"


DEFAULT_DRIVERS = {
    DbType.DB2: "ibm_db_sa",
    DbType.MYSQL: "mysql+pymysql",
    DbType.ORACLE: "oracle+oracledb",
    DbType.PSQL: "postgresql",
    DbType.SQLSERVER: "mssql+pymssql"
}


def get_engine(connection_name: str, context: Context, autocommit: bool = True) -> Tuple[sa.engine.Engine, DbType]:
    connection_details = Connection.get_connection_details(connection_name, context)

    db_type = DbType(connection_details.get("type", "").lower())

    query_args = connection_details.get("query_args", {})
    connect_args = connection_details.get("connect_args", {})
    extra = {}

    if autocommit:
        extra["isolation_level"] = None if db_type == DbType.DB2 else "AUTOCOMMIT"

    if db_type == DbType.ORACLE and connection_details.get("oracle_thick_mode", False):
        lib_dir = connection_details.get("oracle_thick_mode_lib_dir")
        extra["thick_mode"] = {"lib_dir": lib_dir} if lib_dir else {}

    engine = sa.create_engine(
        sa.engine.URL.create(
            drivername=connection_details.get("driver", DEFAULT_DRIVERS.get(db_type)),
            host=connection_details.get("host"),
            port=connection_details.get("port"),
            username=connection_details.get("user"),
            password=connection_details.get("password"),
            database=connection_details.get("database"),
            query=query_args),
        echo=connection_details.get("debug", False),
        connect_args=connect_args,
        **extra)

    return engine, db_type


def construct_table_reference(table: sa.Table, with_brackets: bool = False) -> str:
    """Constructs a table reference string.

    Args:
        table (sa.Table): The SQLAlchemy Table object.
        with_brackets (bool, optional): Whether to include brackets around schema and table names or not.

    Returns:
        str: The formatted table reference string.
    """
    schema_prefix = ""
    if table.schema:
        schema_prefix = f"[{table.schema}]." if with_brackets else f"{table.schema}."
    table_name = f"[{table.name}]" if with_brackets else table.name
    return f"{schema_prefix}{table_name}"



import base64
import hashlib
import re
import string
from datetime import datetime, timezone
from typing import Any, Dict, Iterable, List, Optional, Union
from uuid import uuid4

import jmespath
import orjson
from jmespath import functions


# custom functions for JMESPath
class JmespathCustomFunctions(functions.Functions):

    @functions.signature({"types": ["string", "null"]})
    def _func_capitalize(self, arg):
        return string.capwords(f"{arg}") if arg is not None else None

    @functions.signature({"types": ["array"]})
    def _func_concat(self, elements):
        return "".join([f"{x}" for x in elements])

    @functions.signature({"types": ["string", "null"]})
    def _func_lower(self, element):
        return f"{element}".lower() if element is not None else None

    @functions.signature({"types": ["string", "null"]})
    def _func_upper(self, element):
        return f"{element}".upper() if element is not None else None

    @functions.signature({"types": ["string", "null"]}, {"types": ["string"]}, {"types": ["string"]})
    def _func_replace(self, element, old_value, new_value):
        if element is None:
            return None

        # Unexpected behavior when `old_value` is an empty string in Python's builtin replace
        # https://bugs.python.org/issue28029
        return f"{element}".replace(old_value, new_value) if old_value != "" else f"{element}"

    @functions.signature({"types": ["string", "null"]}, {"types": ["number"]})
    def _func_right(self, element, amount):
        return f"{element}"[-amount:] if element is not None else None

    @functions.signature({"types": ["string", "null"]}, {"types": ["number"]})
    def _func_left(self, element, amount):
        return f"{element}"[:amount] if element is not None else None

    @functions.signature({"types": ["string", "null"]}, {"types": ["number"]}, {"types": ["number"]})
    def _func_mid(self, element, offset, amount):
        return f"{element}"[offset:offset+amount] if element is not None else None

    @functions.signature({"types": ["string", "null"], "variadic": True})
    def _func_split(self, element, delimiter=","):
        return f"{element}".split(delimiter) if element is not None else None

    @functions.signature()
    def _func_uuid(self):
        """Generates a random UUID4 and returns it as a string in standard format."""

        return str(uuid4())

    @functions.signature({"types": ["number", "string", "boolean", "array", "object", "null"], "variadic": True})
    def _func_hash(self, obj, hash_name="sha1"):
        """Calculates a hash using given the `hash_name` hash function and returns its hexadecimal representation.

        Supported algorithms:

        - sha1(default)
        - sha256
        - md5
        - sha384
        - sha3_384
        - blake2b
        - sha512
        - sha3_224
        - sha224
        - sha3_256
        - sha3_512
        - blake2s

        See https://docs.python.org/3/library/hashlib.html for more information.
        """

        def prepare() -> Union[bytes, bytearray]:
            if isinstance(obj, (bytes, bytearray)):
                return obj

            if obj is None:
                return b""

            if isinstance(obj, str):
                return obj.encode()

            return orjson.dumps(obj, option=orjson.OPT_SORT_KEYS)

        h = hashlib.new(hash_name)
        h.update(prepare())
        return h.hexdigest()

    @functions.signature({"types": ["string", "number"]})
    def _func_time_delta_days(self, dt):
        """Returns the number of days between given `dt` and now (positive)
        or the number of days that have passed from now (negative).

        If `dt` is a string, ISO datetime (2011-11-04T00:05:23+04:00, for example) is assumed.
        If `dt` is a number, Unix timestamp (1320365123, for example) is assumed.
        """

        dt = datetime.fromisoformat(dt) if isinstance(dt, str) else datetime.fromtimestamp(dt, timezone.utc)
        delta = dt.now(dt.tzinfo) - dt
        return delta.days

    @functions.signature({"types": ["string", "number"]})
    def _func_time_delta_seconds(self, dt):
        """Returns the number of seconds between given `dt` and now (positive)
        or the number of seconds that have passed from now (negative).

        If `dt` is a string, ISO datetime (2011-11-04T00:05:23+04:00, for example) is assumed.
        If `dt` is a number, Unix timestamp (1320365123, for example) is assumed.
        """

        dt = datetime.fromisoformat(dt) if isinstance(dt, str) else datetime.fromtimestamp(dt, timezone.utc)
        delta = dt.now(dt.tzinfo) - dt

        return delta.days * 86400 + delta.seconds

    @functions.signature({"types": ["string"]}, {"types": ["string"]}, {"types": ["string"]})
    def _func_regex_replace(self, text: str, pattern: str, replacement: str) -> str:
        """Replaces matched patterns in the string by the given replacement."""
        return re.sub(pattern, replacement, text)

    @functions.signature({"types": ["number", "string", "boolean", "array", "object", "null"]}, {"types": ["array"]})
    def _func_in(self, element: Any, iterable: Iterable) -> bool:
        """Returns True if the iterable contains the given element."""
        return element in iterable

    @functions.signature({"types": ["string"]})
    def _func_json_parse(self, data: str) -> Any:
        """Returns parsed object from the given json string."""
        return orjson.loads(data)

    @functions.signature({"types": ["string"]})
    def _func_base64_decode(self, data: str) -> str:
        """Returns decoded string from the given base64 encoded string."""
        return base64.b64decode(data).decode()

    @functions.signature({"types": ["object", "null"]})
    def _func_to_entries(self, obj: Optional[Dict[str, Any]]) -> Optional[List[Dict[str, Any]]]:
        """
        Takes an object and returns an array of {key: key, value: value}.

        Args:
            obj (Optional[Dict[str, Any]]): The object to convert to entries.

        Returns:
            Optional[List[Dict[str, Any]]]: The array of {key: key, value: value} entries.
        """
        if obj is None:
            return None

        return [{"key": key, "value": value} for key, value in obj.items()]

    @functions.signature({"types": ["array", "null"]})
    def _func_from_entries(self, entries: Optional[List[Dict[str, Any]]]) -> Optional[Dict[str, Any]]:
        """
        Takes an array of {key: key, value: value} and returns an object.

        Args:
            entries (Optional[List[Dict[str, Any]]])): The list of entries to convert to an object.

        Returns:
           Optional[Dict[str, Any]]: The object created from the array of entries.
        """
        if entries is None:
            return None

        result = {}
        for entry in entries:
            if isinstance(entry, dict):
                key = entry.get("key")
                if key is not None:
                    result[key] = entry.get("value")

        return result

    @functions.signature({"types": ["object", "null"]}, {"types": ["string"]})
    def _func_filter_entries(self, entries: Optional[Dict[str, Any]],
                             predicate: str) -> Optional[Dict[str, Any]]:
        """
        Filters entries based on the given predicate.

        Args:
            entries (Optional[Dict[str, Any]]): The dictionary (object) of entries to filter.
            predicate (str): The JMESPath predicate to use for filtering.

        Returns:
            Optional[Dict[str, Any]]: The filtered dictionary (object) of entries.
        """
        if entries is None:
            return None

        # Register custom functions
        jmespath_options = jmespath.Options(custom_functions=JmespathCustomFunctions())

        # Compile the predicate expression
        compiled_predicate = jmespath.compile(predicate)

        # Apply the predicate to filter entries
        filtered_entries = {}
        for key, value in entries.items():
            try:
                if compiled_predicate.search({"key": key, "value": value}, options=jmespath_options):
                    filtered_entries[key] = value
            except Exception:
                pass

        return filtered_entries



import logging
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple, Union

from datayoga_core import utils
from datayoga_core.result import Result, Status

logger = logging.getLogger("dy")


def validate_records(records: List[Dict[str, Any]], keys: List[str]) -> Tuple[List[Dict[str, Any]], List[Result]]:

    # validate that the specified keys exist in the records
    rejected_records: List[Result] = []
    valid_records = []
    key_set = set(keys)
    for record in records:
        if key_set - record.keys():
            # not all keys are present. add to rejected
            rejected_records.append(
                Result(status=Status.REJECTED, payload=record, message=f"missing {key_set-record.keys()} field(s)")
            )
        else:
            # all keys are present
            valid_records.append(record)
    return valid_records, rejected_records


def group_records_by_opcode(records: List[Dict[str, Any]],
                            opcode_field: str) -> Dict[str, List[Dict[str, Any]]]:
    # group records by their opcode, reject records with unsupported opcode or with missing keys
    groups = defaultdict(list)
    for record in records:
        groups[record.get(opcode_field, "")].append(record)
    return groups


def get_column_mapping(mapping: List[Union[Dict[str, str], str]]) -> List[Dict[str, str]]:
    return [{"column": next(iter(item.keys())), "key": next(iter(item.values()))} if isinstance(item, dict)
            else {"column": item, "key": item} for item in mapping] if mapping else []


def map_record(record: Dict[str, Any],
               keys: List[Union[Dict[str, str], str]],
               mapping: Optional[List[Union[Dict[str, str], str]]] = None):
    # map the record based on the mapping definitions
    # add nulls for missing mapping fields
    mapping = mapping or []
    mapped_record = {}
    for item in keys + mapping:
        source = next(iter(item.values())) if isinstance(item, dict) else item

        # columns with spaces will be later used with underscores in the bind variables
        target = (next(iter(item.keys())) if isinstance(item, dict) else item).replace(" ", "_")

        source_path = utils.split_field(source)
        obj = record
        for key in source_path:
            key = utils.unescape_field(key)
            if key in obj:
                obj = obj[key]

        mapped_record[target] = obj if obj != record else None

    return mapped_record



import pytest
from datayoga_core import utils
from datayoga_core.blocks.remove_field.block import Block


@pytest.mark.asyncio
async def test_remove_existing_field():
    block = Block({"field": "fname"})
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == utils.all_success([{"lname": "doe"}])


@pytest.mark.asyncio
async def test_remove_missing_field():
    block = Block({"field": "mname"})
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == utils.all_success([{"fname": "john", "lname": "doe"}])


@pytest.mark.asyncio
async def test_remove_deep_nested_field():
    block = Block({"field": "employee.name.fname"})
    block.init()
    assert await block.run([{"employee": {"name": {"fname": "john", "lname": "doe"}}}]) == utils.all_success([
        {"employee": {"name": {"lname": "doe"}}}])


@pytest.mark.asyncio
async def test_remove_nested_field():
    block = Block({"field": "name.fname"})
    block.init()
    assert await block.run([{"name": {"fname": "john", "lname": "doe"}}]) == utils.all_success([{"name": {"lname": "doe"}}])


@pytest.mark.asyncio
async def test_remove_missing_nested_field():
    block = Block({"field": "full_name.fname"})
    block.init()
    assert await block.run([{"name": {"fname": "john", "lname": "doe"}}]) == utils.all_success([{"name": {"fname": "john", "lname": "doe"}}])


@pytest.mark.asyncio
async def test_remove_multiple_fields():
    block = Block({"fields": [{"field": "name.fname"}, {"field": "name.lname"}]})
    block.init()
    assert await block.run([{"name": {"fname": "john", "lname": "doe"}}]) == utils.all_success([{"name": {}}])


@pytest.mark.asyncio
async def test_remove_field_with_dot():
    block = Block({"field": "name\.fname"})
    block.init()
    assert await block.run([{"name.fname": "john", "lname": "doe"}]) == utils.all_success([{"lname": "doe"}])



from __future__ import annotations

import asyncio
import logging
from asyncio import Task
from typing import Any, Callable, Dict, List, Optional

from datayoga_core import prometheus
from datayoga_core.block import Block
from datayoga_core.context import Context
from datayoga_core.result import Result, Status

logger = logging.getLogger("dy")


class Step:
    def __init__(self, step_id: str, block: Optional[Block], concurrency=1):
        self.id = step_id
        self.block = block
        self.next_step = None
        self.active_entries = set()
        self.concurrency = concurrency
        self.workers: List[Optional[Task]] = [None]*self.concurrency
        self.done_callback = None
        self.initialized = False

    def init(self, context: Optional[Context] = None):
        # initialize the block
        self.block.init(context)

    async def start_pool(self):
        # start pool of workers for parallelization
        logger.debug("starting pool")
        self.queue = asyncio.Queue(maxsize=1)
        for worker_id in range(self.concurrency):
            worker = self.workers[worker_id]
            if worker is None or not worker.done():
                self.workers[worker_id] = asyncio.create_task(self.run(worker_id))
            else:
                logger.debug(f"worker {worker_id} is running: {not worker.done()}")

    def add_done_callback(self, callback: Callable[[List[str], List[Result]], None]):
        self.done_callback = callback

    def __or__(self, other: Step):
        return self.append(other)

    def append(self, next_step: Step):
        self.next_step = next_step
        self.next_step.add_done_callback(self.done)
        return self.next_step

    async def process(self, messages: List[Dict[str, Any]]):
        if not self.initialized:
            await self.start_pool()
            self.initialized = True
        self.active_entries.update([x[Block.MSG_ID_FIELD] for x in messages])
        await self.queue.put(messages)

    async def run(self, worker_id: int):
        while True:
            entry = await self.queue.get()
            logger.debug(f"{self.id}-{worker_id} processing {[i[Block.MSG_ID_FIELD] for i in entry]}")
            try:
                processed_entries, filtered_entries, rejected_entries = await self.block.run(entry)

                prometheus.processed_entries.labels(step=self.id).inc(len(processed_entries))
                prometheus.filtered_records.labels(step=self.id).inc(len(filtered_entries))
                prometheus.rejected_records.labels(step=self.id).inc(len(rejected_entries))

                # handle filtered. anything not processed or rejected
                logger.debug(
                    f"filtered entries: {filtered_entries}, processed entries: {processed_entries}, rejected entries: {rejected_entries}")
                if filtered_entries:
                    # ack the filtered entries
                    self.done([row.payload[Block.MSG_ID_FIELD] for row in filtered_entries],
                              [Result(Status.FILTERED)] * len(filtered_entries))

                # handle rejected
                if rejected_entries:
                    # ack the rejected entries
                    self.done([row.payload[Block.MSG_ID_FIELD] for row in rejected_entries], rejected_entries)

                if processed_entries:
                    # check if we have a next step
                    if self.next_step:
                        # process downstream
                        await self.next_step.process([result.payload for result in processed_entries])
                    else:
                        # we are a last channel, propagate the ack upstream
                        # TODO: verify that all entries have a msg id otherwise raise consistency error
                        self.done([row.payload[Block.MSG_ID_FIELD] for row in processed_entries], processed_entries)

            except Exception as e:
                # we caught an exception. the entire batch is considered rejected
                logger.exception(e)
                # verify that all messages still have a msg_id property
                # if next(filter(lambda x: not Block.MSG_ID_FIELD in x,entry),None) is not None
                self.done([x[Block.MSG_ID_FIELD] for x in entry], [
                          Result(Status.REJECTED, f"Error in step {self.id}: {repr(e)}")] * len(entry))
            finally:
                self.queue.task_done()
            logger.debug(f"{self.id}-{worker_id} done processing {entry}")

    def done(self, msg_ids: List[str], results: List[Result]):
        logger.debug(f"{self.id} acking {msg_ids}")
        self.active_entries.difference_update(msg_ids)
        if self.done_callback is not None:
            self.done_callback(msg_ids, results)

    async def join(self):
        # wait for all active entries to be processed
        while len(self.active_entries) > 0:
            logger.debug(f"{self.id} waiting for dangling messages: {self.active_entries}")
            await asyncio.sleep(0.2)

    async def stop(self):
        # wait for all tasks to finish
        await self.join()
        logger.debug("joined")

        # stop the block
        if self.block:
            self.block.stop()

        # stop any downstream workers
        if self.next_step:
            await self.next_step.stop()

        # stop all workers in the pool
        for worker in self.workers:
            if worker:
                worker.cancel()



import pytest
from datayoga_core import utils
from datayoga_core.blocks.map.block import Block


@pytest.mark.asyncio
async def test_map_expression_jmespath():
    block = Block(properties={"language": "jmespath",
                              "expression": """{
                        "new_field": `hello`
                   }"""})
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == utils.all_success([
        {"new_field": "hello"}]
    )


@pytest.mark.asyncio
async def test_map_field_multiple_expressions_jmespath():
    block = Block(properties={"language": "jmespath",
                              "expression": """{
                            "new_field": `hello`, "name" : fname
                    }"""})
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == utils.all_success([
        {"new_field": "hello", "name": "john"}]
    )


@pytest.mark.asyncio
async def test_map_field_nested_expression_jmespath():
    block = Block(properties={"language": "jmespath",
                              "expression": """{
                            "new_field": `hello`, "name" : details.fname
                    }"""})
    block.init()
    assert await block.run([{"details": {"fname": "john", "lname": "doe"}}]) == utils.all_success([
        {"new_field": "hello", "name": "john"}]
    )


@pytest.mark.asyncio
async def test_map_field_double_nested_expression_jmespath():
    block = Block(properties={"language": "jmespath",
                              "expression": """{
                            "new_field": `hello`, "name" : details.name.fname
                    }"""})
    block.init()
    assert await block.run([{"details": {"name": {"fname": "john", "lname": "doe"}, "country": "israel"}}]) == utils.all_success([
        {"new_field": "hello", "name": "john"}]
    )


@pytest.mark.asyncio
async def test_map_field_nested_expression_jmespath_object():
    block = Block(properties={"language": "jmespath", "expression": {"new_field": "`hello`", "name": "details.fname"}})
    block.init()
    assert await block.run([{"details": {"fname": "john", "lname": "doe"}}]) == utils.all_success([
        {"new_field": "hello", "name": "john"}]
    )


@pytest.mark.asyncio
async def test_map_field_double_nested_expression_jmespath_object():
    block = Block(properties={"language": "jmespath", "expression": {
                  "new_field": "`hello`", "name": "details.name.fname"}})
    block.init()
    assert await block.run([{"details": {"name": {"fname": "john", "lname": "doe"}, "country": "israel"}}]) == utils.all_success([
        {"new_field": "hello", "name": "john"}]
    )


@pytest.mark.asyncio
async def test_map_expression_non_quoted_jmespath():
    block = Block(properties={"language": "jmespath",
                              "expression": {
                                  "name": "fname"
                              }})
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == utils.all_success([{"name": "john"}])


@pytest.mark.asyncio
async def test_map_multiple_expressions_non_quoted_jmespath():
    block = Block(properties={"language": "jmespath",
                              "expression": {
                                  "name": "fname", "last name": "lname"
                              }})
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == utils.all_success([
        {"name": "john", "last name": "doe"}]
    )


@pytest.mark.asyncio
async def test_map_expression_sql():
    block = Block(properties={"language": "sql",
                              "expression": {
                                  "new_field": "fname"
                              }})
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == utils.all_success([
        {"new_field": "john"}]
    )


@pytest.mark.asyncio
async def test_map_multiple_expressions_sql():
    block = Block(properties={"language": "sql",
                              "expression": {
                                  "name": "fname", "last name": "lname"
                              }})
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == utils.all_success([
        {"name": "john", "last name": "doe"}]
    )


@pytest.mark.asyncio
async def test_map_double_nested_expression_sql():
    block = Block(properties={"language": "sql",
                              "expression": {
                                  "name": "(`details.name.fname`)"
                              }})
    block.init()
    assert await block.run([{"details": {"name": {"fname": "john", "lname": "doe"}, "country": "israel"}}]) == utils.all_success([
        {"name": "john"}]
    )


@pytest.mark.asyncio
async def test_map_nested_expression_sql():
    block = Block(properties={"language": "sql",
                              "expression": {
                                  "name": "(`details.fname`)"
                              }})
    block.init()
    assert await block.run([{"details": {"fname": "john", "lname": "doe"}}]) == utils.all_success([
        {"name": "john"}]
    )


@pytest.mark.asyncio
async def test_map_malformed():
    block = Block(properties={"language": "sql",
                              "expression": "{name: (`details.fname`) "})
    with pytest.raises(ValueError):
        block.init()


@pytest.mark.asyncio
async def test_jmespath_does_not_filter_in_map():
    block = Block(properties={"language": "jmespath",
                              "expression": "{name: a} "})
    block.init()
    assert await block.run([{"a": "one"}, {"b": "two"}]) == utils.all_success([
        {"name": "one"},
        {"name": None}
    ])


@pytest.mark.asyncio
async def test_map_malformed_sql():
    block = Block(properties={"language": "sql",
                              "expression": "{name: (`details.fname`) "})
    with pytest.raises(ValueError):
        block.init()
        await block.run([{"details": {"fname": "john", "lname": "doe"}}])



import logging
import os
from abc import ABCMeta
from typing import AsyncGenerator, List, Optional

from datayoga_core.context import Context
from datayoga_core.producer import Message
from datayoga_core.producer import Producer as DyProducer
from fastparquet import ParquetFile

logger = logging.getLogger("dy")


class Block(DyProducer, metaclass=ABCMeta):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        parquet_file = self.properties["file"]

        if os.path.isabs(parquet_file) or context is None:
            self.file = parquet_file
        else:
            self.file = os.path.join(context.properties.get("data_path"), parquet_file)

        logger.debug(f"file: {self.file}")

    async def produce(self) -> AsyncGenerator[List[Message], None]:
        logger.debug("Reading parquet")

        pf = ParquetFile(self.file)

        count = 0
        for df in pf.iter_row_groups():
            for _, data in df.iterrows():
                yield [{self.MSG_ID_FIELD: str(count), **data.to_dict()}]
                count += 1



import logging
from typing import AsyncGenerator, List, Optional

import datayoga_core.blocks.redis.utils as redis_utils
import orjson
from datayoga_core.connection import Connection
from datayoga_core.context import Context
from datayoga_core.producer import Message
from datayoga_core.producer import Producer as DyProducer

logger = logging.getLogger("dy")


class Block(DyProducer):
    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")

        connection_details = Connection.get_connection_details(self.properties["connection"], context)
        self.redis_client = redis_utils.get_client(connection_details)

        self.stream = self.properties["stream_name"]
        self.snapshot = self.properties.get("snapshot", False)
        self.consumer_group = f'datayoga_job_{context.properties.get("job_name", "") if context else ""}'
        self.requesting_consumer = "dy_consumer_a"
        stream_groups = self.redis_client.xinfo_groups(self.stream)
        if next(filter(lambda x: x["name"] == self.consumer_group, stream_groups), None) is None:
            logger.info(f"Creating a new {self.consumer_group} consumer group associated with the {self.stream}")
            self.redis_client.xgroup_create(self.stream, self.consumer_group, 0)

    async def produce(self) -> AsyncGenerator[List[Message], None]:
        logger.debug(f"Running {self.get_block_name()}")

        read_pending = True
        while True:
            # Read pending messages (fetched by us before but not acknowledged) in the first time, then consume new messages
            streams = self.redis_client.xreadgroup(self.consumer_group, self.requesting_consumer, {
                self.stream: "0" if read_pending else ">"}, None, 100 if self.snapshot else 0)

            for stream in streams:
                logger.debug(f"Messages in {self.stream} stream (pending: {read_pending}):\n\t{stream}")
                for key, value in stream[1]:
                    payload = orjson.loads(value[next(iter(value))])
                    payload[self.MSG_ID_FIELD] = key
                    yield [payload]

            # Quit after consuming pending current messages in case of snapshot
            if self.snapshot and not read_pending:
                break

            read_pending = False

    def ack(self, msg_ids: List[str]):
        for msg_id in msg_ids:
            logger.info(f"Acking {msg_id} message in {self.stream} stream of {self.consumer_group} consumer group")
            self.redis_client.xack(self.stream, self.consumer_group, msg_id)



import logging
from abc import ABCMeta
from typing import Any, Dict, List, Optional

from datayoga_core.block import Block as DyBlock
from datayoga_core.context import Context

logger = logging.getLogger(__name__)


class Block(DyBlock, metaclass=ABCMeta):

    def init(self, context: Optional[Context] = None):
        pass

    def run(self, data: List[Dict[str, Any]], context: Context = None) -> List[Dict[str, Any]]:
        pass



import os
from pathlib import Path
from typing import Any, Dict

from datayoga_core import utils
from datayoga_core.context import Context


class Connection:
    """Connection"""

    @staticmethod
    def get_connection_details(connection_name: str, context: Context) -> Dict[str, Any]:
        """Gets connection details from the context"""
        if context and context.properties:
            connection = context.properties.get("connections", {}).get(connection_name)
            if connection:
                return connection

        raise ValueError(f"{connection_name} connection not found")

    @staticmethod
    def get_json_schema() -> Dict[str, Any]:
        """Compiles a complete JSON schema of the connection with all possible types"""
        connection_schemas = []

        connections_dir = utils.get_resource_path(os.path.join("schemas", "connections"))

        for schema_path in sorted(Path(connections_dir).rglob("**/*.schema.json"), key=lambda p: p.stem):
            connection_schemas.append(utils.read_json(f"{schema_path}"))

        connections_general_schema = utils.read_json(
            os.path.join(
                utils.get_bundled_dir() if utils.is_bundled() else os.path.dirname(os.path.realpath(__file__)),
                "resources", "schemas", "connections.schema.json"))

        connections_general_schema["definitions"]["connection"]["oneOf"] = connection_schemas

        return connections_general_schema
